@@ -30,6 +30,7 @@ import shlex
 import signal
 import subprocess
 import tempfile
+import textwrap
 import shutil
 from string import Template
 import urllib.parse
@@ -61,7 +62,7 @@ class Testbed:
                  copy_files=[], pin_packages=[], add_apt_sources=[],
                  add_apt_releases=[], apt_default_release=None,
                  enable_apt_fallback=True, shell_fail=False, needs_internet='run',
-                 foreign_test_arch=None):
+                 test_arch=None):
         self.sp = None
         self.lastsend = None
         self.scratch = None
@@ -69,7 +70,7 @@ class Testbed:
         self._need_reset_apt = False
         self.stop_sent = False
         self.dpkg_arch = None
-        self.foreign_test_arch = foreign_test_arch
+        self.test_arch = test_arch
         self.cross_env = []
         self.exec_cmd = None
         self.output_dir = output_dir
@@ -318,28 +319,30 @@ class Testbed:
         self.dpkg_arch = self.check_exec(['dpkg', '--print-architecture'], True).strip()
         adtlog.info('testbed dpkg architecture: ' + self.dpkg_arch)
 
-        if self.foreign_test_arch and self.foreign_test_arch == self.dpkg_arch:
-            adtlog.info(
-                f'ignoring the requested test architecture ({self.foreign_test_arch}) '
-                'as it matches the testbed native dpkg architecture.'
-            )
-            self.foreign_test_arch = None
+        # the test architesture defaults to the dpkg architecture
+        if not self.test_arch:
+            self.test_arch = self.dpkg_arch
+
+        self.test_arch_is_foreign = (self.test_arch != self.dpkg_arch)
 
         out = self.check_exec(['dpkg-query', '-W', '-f', '${Version}', 'apt'], True).strip()
         adtlog.info('testbed apt version: ' + out)
         try:
             self.apt_version = Version(out)
         except ValueError:
-            self.apt_version = None
+            self.bomb("can't determine the apt version on the testbed")
+
+        # Foreign arch testing requires `apt-get satisfy`.
+        if self.test_arch_is_foreign and self.apt_version < Version("1.9.0"):
+            self.bomb("foreign arch testing requires apt >= 1.9.0 on the testbed")
 
         # set up environment for cross-architecture if needed
-        if self.foreign_test_arch:
-            argv = ['dpkg', '--add-architecture', self.foreign_test_arch]
+        if self.test_arch_is_foreign:
+            argv = ['dpkg', '--add-architecture', self.test_arch]
             self.check_exec(argv)
 
-            self.cross_env = []
-            # dpkg-architecture (<< 1.17.17) only supports the -a<arch> syntax
-            argv = ['dpkg-architecture', '-a' + self.foreign_test_arch]
+            # dpkg-architecture (<< 1.17.17) only supports the short -a<arch> syntax
+            argv = ['dpkg-architecture', '-a' + self.test_arch]
             # dpkg-architectures prints warnings to stderr if gcc is missing,
             # do don't fail on stderr.
             vars = self.check_exec(argv, stdout=True, fail_on_stderr=False)
@@ -349,7 +352,7 @@ class Testbed:
                 # scripts with both the --build and --host arguments.
                 if var.startswith('DEB_HOST_') or var.startswith('DEB_BUILD_'):
                     self.cross_env.append(var)
-            adtlog.info('test architecture: ' + self.foreign_test_arch)
+            adtlog.info('test architecture: ' + self.test_arch)
 
         self.run_setup_commands()
 
@@ -538,31 +541,22 @@ class Testbed:
             self._opened(pl)
         self.modified = False
 
-    def install_deps(self, deps_new, shell_on_failure=False, synth_deps=[],
-                     foreign_test_arch=None):
+    def install_deps(self, deps_new, shell_on_failure=False, synth_deps=[]):
         '''Install dependencies into testbed'''
         adtlog.debug('install_deps: deps_new=%s' % deps_new)
 
         self.deps_installed = deps_new
         if not deps_new:
             return
-
-        if self.foreign_test_arch:
-            # mock up a .dsc with our dependencies string as our build-deps
-            dsc = TempPath(self, 'autopkgtest-satdep.dsc')
-            with open(dsc.host, 'w', encoding='UTF-8') as f_dsc:
-                f_dsc.write("Source: autopkgtest-satdep\n"
-                            "Binary: autopkgtest-satdep\n"
-                            "Architecture: any all\n"
-                            "Build-Depends: " + ', '.join(deps_new) + "\n")
-            dsc.copydown()
-
-            # feed the result to apt-get build-dep
-            self.install_build_deps_for_package(dsc.tb, shell_on_failure=shell_on_failure)
-            # Now handle the synthetic deps as well
-            self.install_apt('', shell_on_failure, synth_deps)
-        else:
-            self.satisfy_dependencies_string(', '.join(deps_new), 'install-deps', shell_on_failure=shell_on_failure, synth_deps=synth_deps)
+        # Test dependencies are treated as build dependencies, which in
+        # practice means that the :native architecture qualifier is supported.
+        self.satisfy_dependencies_string(
+            ', '.join(deps_new),
+            'install-deps',
+            build_dep=True,
+            shell_on_failure=shell_on_failure,
+            synth_deps=synth_deps
+        )
 
     def _provide_sudo(self) -> str:
         '''
@@ -872,8 +866,8 @@ class Testbed:
                     return True
         return False
 
-    def _run_apt_install(self, what, prefix, action='install', extra_args=None, ignorerc=False, xenv=[]):
-        '''call apt-get with an install-like action (e.g. install, build-dep)'''
+    def _run_apt_install(self, what, prefix, action='install', extra_args=None, xenv=[]):
+        '''call apt-get with an install-like action (e.g. install, build-dep, satisfy)'''
 
         cmd = [
             prefix,
@@ -894,7 +888,7 @@ class Testbed:
         # capture status-fd to stderr
         (rc, _, serr) = self.execute(['/bin/sh', '-ec', ' '.join(cmd)],
                                      kind='install', stderr=subprocess.PIPE, xenv=xenv)
-        if not ignorerc and rc != 0:
+        if rc != 0:
             adtlog.debug('apt-get %s %s failed; status-fd:\n%s' % (action, what, serr))
             # check if apt failed during package download, which might be a
             # transient error, so retry
@@ -938,53 +932,90 @@ class Testbed:
             adtlog.debug('installing real packages of test dependencies: %s' % need_explicit_install)
             self._run_apt_install(' '.join(need_explicit_install), ' '.join(self.eatmydata_prefix))
 
-    def install_apt(self, deps, shell_on_failure=False, synth_deps=[]):
-        '''Install dependencies with apt-get into testbed
-
-        This requires root privileges and a writable file system.
-        '''
-        # create a dummy deb with the deps
+    def _generate_satdep_deb(self, deps):
+        '''create a dummy deb with the deps'''
         pkgdir = tempfile.mkdtemp(prefix='autopkgtest-satdep.')
         debdir = os.path.join(pkgdir, 'DEBIAN')
         os.chmod(pkgdir, 0o755)
         os.mkdir(debdir)
         os.chmod(debdir, 0o755)
         with open(os.path.join(debdir, 'control'), 'w') as f:
-            f.write('''Package: autopkgtest-satdep
-Section: oldlibs
-Priority: extra
-Maintainer: autogenerated
-Version: 0
-Architecture: %s
-Depends: %s
-Description: satisfy autopkgtest test dependencies
-''' % (self.dpkg_arch, deps))
-
+            f.write(
+                textwrap.dedent(
+                    f"""\
+                    Package: autopkgtest-satdep
+                    Section: oldlibs
+                    Priority: extra
+                    Maintainer: autogenerated
+                    Version: 0
+                    Architecture: {self.dpkg_arch}
+                    Depends: {deps}
+                    Description: satisfy autopkgtest test dependencies
+                    """
+                )
+            )
         deb = TempPath(self, 'autopkgtest-satdep.deb')
         subprocess.check_call(['dpkg-deb', '-Zxz', '-b', pkgdir, deb.host],
                               stdout=subprocess.PIPE)
         shutil.rmtree(pkgdir)
-        deb.copydown()
+        return deb
 
-        apt_version = self.apt_version
-        assert apt_version is not None
-        # Debian >= 9, Ubuntu >= 16.04
-        apt_can_install_debs = (apt_version >= Version('1.1~exp1'))
+    def install_apt(self, deps, shell_on_failure=False, synth_deps=[]):
+        '''Install dependencies with apt-get into testbed
 
-        # install it and its dependencies in the tb; our apt pinning is not
-        # very clever wrt. resolving transitional dependencies in the pocket,
+        This requires root privileges and a writable file system.
+        '''
+        # Check if apt supports the `apt-get satisfy <string>` syntax
+        legacy_apt = (self.apt_version < Version('1.9.0'))
+
+        if legacy_apt:
+            # Fallback to installing test dependencies via a dummy binary .deb.
+            #
+            # TODO for when we'll be able to assume apt (>= 1.1exp1) on the testbed:
+            # 1. Switch from installing deps via `dpkg --unpack autopkgtest-satdep.deb`
+            #    + `apt-get --fix-broken` + `apt-get purge autokgtest-satdep` to
+            #    `apt-get build-dep autopkgtest-satdep.dsc`.
+            #    See: b86576aa47f6f95985c3be46e4a4e20278da4511.
+            # 2. Drop the special handling of :native.
+            adtlog.debug('Legacy testbed apt, installing dependencies via autopkgtest-satdep.deb')
+
+            # We need to strip :native qualifiers as they are not valid in
+            # DEBIAN/control Depends, see deb-control(5), and compare with
+            # deb-src-control(5).
+            deps = deps.replace(':native', '')
+
+            try:
+                deps = subprocess.run(
+                    [
+                        os.path.join(PKGDATADIR, 'lib', 'parse-deps.pl'),
+                        deps,
+                        self.dpkg_arch,
+                    ],
+                    check=True,
+                    capture_output=True,
+                    text=True,
+                ).stdout.strip()
+            except subprocess.CalledProcessError as e:
+                self.bomb(f'failed to run: {e.cmd}')
+            adtlog.debug(f"architecture resolved deps: {deps}")
+
+            # generate a dummy package with the deps
+            satdep_package = self._generate_satdep_deb(deps)
+            satdep_package.copydown()
+
+        # install the dependencies in the tb; our apt pinning is not very
+        # clever wrt. resolving transitional dependencies in the pocket,
         # so we might need to retry without pinning
         while True:
-            if not apt_can_install_debs:
-                self.check_exec(['dpkg', '--unpack', deb.tb], stdout=subprocess.PIPE)
+            if legacy_apt:
+                self.check_exec(['dpkg', '--unpack', satdep_package.tb], stdout=subprocess.PIPE)
 
             rc = 0
             try:
-                if apt_can_install_debs:
-                    self._run_apt_install(deb.tb, ' '.join(self.eatmydata_prefix))
-                else:
+                if legacy_apt:
                     self._run_apt_install('--fix-broken', ' '.join(self.eatmydata_prefix))
-
+                elif deps:
+                    self._run_apt_install(f"'{deps}'", ' '.join(self.eatmydata_prefix), action='satisfy')
                 # if '@' was provided as a test dependency (as seen in synth_deps),
                 # we explicitly install the corresponding binary packages in case
                 # the dependencies on those were satisfied by versioned Provides
@@ -998,10 +1029,11 @@ Description: satisfy autopkgtest test dependencies
                 if shell_on_failure:
                     self.run_shell()
             else:
-                # verify that we actually got autopkgtest-satdep installed
-                rc = self.execute(['dpkg', '--status', 'autopkgtest-satdep'],
-                                  stdout=subprocess.PIPE,
-                                  stderr=subprocess.PIPE)[0]
+                if legacy_apt:
+                    # verify that we actually got autopkgtest-satdep installed
+                    rc = self.execute(['dpkg', '--status', 'autopkgtest-satdep'],
+                                      stdout=subprocess.PIPE,
+                                      stderr=subprocess.PIPE)[0]
 
             if rc != 0:
                 if self.apt_pin_for_releases and self.enable_apt_fallback:
@@ -1013,13 +1045,6 @@ Description: satisfy autopkgtest test dependencies
                         self.check_exec(['/bin/sh', '-ec', 'rm -f /etc/apt/preferences.d/autopkgtest-default-release'])
                     continue
 
-                adtlog.warning('Test dependencies are unsatisfiable - calling '
-                               'apt install on test deps directly for further '
-                               'data about failing dependencies in test logs')
-                self._run_apt_install('--simulate ' + deps.replace(',', ''),
-                                      ' '.join(self.eatmydata_prefix),
-                                      ignorerc=True)
-
                 if shell_on_failure:
                     self.run_shell()
                 if self.enable_apt_fallback:
@@ -1035,30 +1060,31 @@ Description: satisfy autopkgtest test dependencies
                                 'list.')
             break
 
-        # remove autopkgtest-satdep to avoid confusing tests, but avoid marking our
-        # test dependencies for auto-removal
-        out = self.check_exec(['apt-get', '--simulate', '--quiet',
-                               '-o', 'APT::Get::Show-User-Simulation-Note=False',
-                               '--auto-remove',
-                               'purge', 'autopkgtest-satdep'],
-                              True)
-        test_deps = []
-        for line in out.splitlines():
-            if not line.startswith('Purg '):
-                continue
-            pkg = line.split()[1]
-            if pkg != 'autopkgtest-satdep':
-                test_deps.append(pkg)
-        if test_deps:
-            adtlog.debug('Marking test dependencies as manually installed: %s' %
-                         ' '.join(test_deps))
-            # avoid overly long command lines
-            batch = 0
-            while batch < len(test_deps):
-                self.check_exec(['apt-mark', 'manual', '-qq'] + test_deps[batch:batch + 20])
-                batch += 20
+        if legacy_apt:
+            # remove autopkgtest-satdep to avoid confusing tests, but avoid marking our
+            # test dependencies for auto-removal
+            out = self.check_exec(['apt-get', '--simulate', '--quiet',
+                                   '-o', 'APT::Get::Show-User-Simulation-Note=False',
+                                   '--auto-remove',
+                                   'purge', 'autopkgtest-satdep'],
+                                  True)
+            test_deps = []
+            for line in out.splitlines():
+                if not line.startswith('Purg '):
+                    continue
+                pkg = line.split()[1]
+                if pkg != 'autopkgtest-satdep':
+                    test_deps.append(pkg)
+            if test_deps:
+                adtlog.debug('Marking test dependencies as manually installed: %s' %
+                             ' '.join(test_deps))
+                # avoid overly long command lines
+                batch = 0
+                while batch < len(test_deps):
+                    self.check_exec(['apt-mark', 'manual', '-qq'] + test_deps[batch:batch + 20])
+                    batch += 20
 
-        self.execute(['dpkg', '--purge', 'autopkgtest-satdep'])
+            self.execute(['dpkg', '--purge', 'autopkgtest-satdep'])
 
     def install_build_deps_for_package(self, src_package_path, shell_on_failure=False, xenv=[]) -> None:
         '''Install build dependencies via apt-get build-dep.
@@ -1091,7 +1117,7 @@ Description: satisfy autopkgtest test dependencies
         needed_bd = [line for line in bd_simulate_out if re.match(r'^Inst ', line)]
 
         if not needed_bd:
-            adtlog.debug('No build depdendencies to install')
+            adtlog.debug('No build dependencies to install')
             return
 
         if 'root-on-testbed' not in self.caps:
@@ -1120,38 +1146,21 @@ Description: satisfy autopkgtest test dependencies
 
         adtlog.debug('%s: satisfying %s' % (what, deps))
 
-        # ignore ":native" tags, apt cannot parse them and deps_parse() does
-        # not seem to have an option to get rid of them; we always test on the
-        # native platform
-        deps = deps.replace(':native', '')
-
-        # resolve arch specific dependencies; don't use universal_newlines
-        # here, it's broken for stdin on Python 3.2
-        subproc = subprocess.Popen(
-            [
-                os.path.join(PKGDATADIR, 'lib', 'parse-deps.pl'),
-                deps,
-                self.dpkg_arch,
-                'build-depends' if build_dep else 'depends',
-            ],
-            stdout=subprocess.PIPE,
-        )
-        deps = subproc.communicate()[0].decode('UTF-8').strip()
-        if subproc.returncode != 0:
-            self.bomb(f'failed to run {subproc.args[0]}')
-        adtlog.debug('%s: architecture resolved: %s' % (what, deps))
-
         # check if we can use apt-get
         can_apt_get = False
-        if 'root-on-testbed' in self.caps and self.apt_version is not None:
-            rc = self.execute(['test', '-w', '/var/lib/dpkg/status'])[0]
-            if rc == 0:
-                can_apt_get = True
+        if 'root-on-testbed' in self.caps:
+            can_apt_get = True
         adtlog.debug('can use apt-get on testbed: %s' % can_apt_get)
 
         if can_apt_get:
             self.install_apt(deps, shell_on_failure, synth_deps)
         else:
+            # This whole `else` branch is meant to check if we can get
+            # away without being root on the testbed because all the
+            # dependencies are already installed. Hopefully one day
+            # we'll replace it with something better, e.g. checking
+            # the output of `apt-get --simulate`.
+
             has_dpkg_checkbuilddeps = self.execute(
                 ['sh', '-ec', 'command -v dpkg-checkbuilddeps'],
                 stdout=subprocess.DEVNULL,
@@ -1232,9 +1241,7 @@ Description: satisfy autopkgtest test dependencies
             '--artifacts={}'.format(test_artifacts),
             '--chdir={}'.format(tree.tb),
             '--env=AUTOPKGTEST_TESTBED_ARCH={}'.format(self.dpkg_arch),
-            '--env=AUTOPKGTEST_TEST_ARCH={}'.format(
-                self.foreign_test_arch if self.foreign_test_arch else self.dpkg_arch
-            ),
+            '--env=AUTOPKGTEST_TEST_ARCH={}'.format(self.test_arch),
             '--env=DEB_BUILD_OPTIONS=parallel={}'.format(
                 build_parallel or self.nproc
             ),
@@ -1441,7 +1448,7 @@ Description: satisfy autopkgtest test dependencies
             self.run_shell(tree.tb, ['AUTOPKGTEST_ARTIFACTS="%s"' % test_artifacts,
                                      'AUTOPKGTEST_TMP="%s"' % autopkgtest_tmp,
                                      'AUTOPKGTEST_TESTBED_ARCH=%s' % self.dpkg_arch,
-                                     'AUTOPKGTEST_TEST_ARCH=%s' % self.foreign_test_arch if self.foreign_test_arch else self.dpkg_arch])
+                                     'AUTOPKGTEST_TEST_ARCH=%s' % self.test_arch])
 
         # clean up artifacts and AUTOPKGTEST_TMP dirs
         self.check_exec(['rm', '-rf', test_artifacts, autopkgtest_tmp])
@@ -1485,9 +1492,7 @@ Pin: release a=$release
 Pin-Priority: $prio
 
 """)
-        apt_version = self.apt_version
-        assert apt_version is not None
-        if apt_version < Version('2.0'):
+        if self.apt_version < Version('2.0'):
             # sort pkglist into source and binary packages
             binpkgs = []
             srcpkgs = []
@@ -1499,6 +1504,10 @@ Pin-Priority: $prio
                     binpkgs.append(i)
 
             # translate src:name entries into binaries of that source
+            #
+            # apt 1.9.11 implemented support for pinning by source package
+            # using the src:pkg syntax, as documented in apt_preferences(5),
+            # however that's too new for us to use for now.
             if srcpkgs:
                 apt_cache = self.check_exec(
                     ['apt-cache', 'dumpavail'], True)
@@ -1515,10 +1524,12 @@ Pin-Priority: $prio
         # prefer given packages from source suite, but make sure that
         # other packages are taken from default suite as much as
         # possible
-        pref_content = pref_template.substitute(
-            pkgs=pkglist.replace(',', ' '),
-            release=release,
-            prio='995')
+        pref_content = ""
+        if pkglist:
+            pref_content += pref_template.substitute(
+                pkgs=pkglist.replace(',', ' '),
+                release=release,
+                prio='995')
         for default in default_releases:
             pref_content += pref_template.substitute(
                 pkgs="*",
